@@ -6,8 +6,6 @@
  * a multiplier factor from there, up to half the maximum slab size. The last
  * slab size is always 1MB, since that's the maximum item size allowed by the
  * memcached protocol.
- *
- * $Id: slabs.c 745 2008-03-03 07:51:02Z dsallings $
  */
 #include "memcached.h"
 #include <sys/stat.h>
@@ -21,12 +19,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <assert.h>
-
-#define POWER_SMALLEST 1
-#define POWER_LARGEST  200
-#define POWER_BLOCK 1048576
-#define CHUNK_ALIGN_BYTES 8
-#define DONT_PREALLOC_SLABS
+#include <pthread.h>
 
 /* powers-of-N allocation structures */
 
@@ -47,9 +40,10 @@ typedef struct {
     unsigned int list_size; /* size of prev array */
 
     unsigned int killing;  /* index+1 of dying slab, or zero if none */
+    size_t requested; /* The number of requested bytes */
 } slabclass_t;
 
-static slabclass_t slabclass[POWER_LARGEST + 1];
+static slabclass_t slabclass[MAX_NUMBER_OF_SLAB_CLASSES];
 static size_t mem_limit = 0;
 static size_t mem_malloced = 0;
 static int power_largest;
@@ -58,6 +52,11 @@ static void *mem_base = NULL;
 static void *mem_current = NULL;
 static size_t mem_avail = 0;
 
+/**
+ * Access to the slab allocator is protected by this lock
+ */
+static pthread_mutex_t slabs_lock = PTHREAD_MUTEX_INITIALIZER;
+
 /*
  * Forward Declarations
  */
@@ -121,23 +120,27 @@ void slabs_init(const size_t limit, const double factor, const bool prealloc) {
 
     memset(slabclass, 0, sizeof(slabclass));
 
-    while (++i < POWER_LARGEST && size <= POWER_BLOCK / 2) {
+    while (++i < POWER_LARGEST && size <= settings.item_size_max / 2) {
         /* Make sure items are always n-byte aligned */
         if (size % CHUNK_ALIGN_BYTES)
             size += CHUNK_ALIGN_BYTES - (size % CHUNK_ALIGN_BYTES);
 
         slabclass[i].size = size;
-        slabclass[i].perslab = POWER_BLOCK / slabclass[i].size;
+        slabclass[i].perslab = settings.item_size_max / slabclass[i].size;
         size *= factor;
         if (settings.verbose > 1) {
-            fprintf(stderr, "slab class %3d: chunk size %6u perslab %5u\n",
+            fprintf(stderr, "slab class %3d: chunk size %9u perslab %7u\n",
                     i, slabclass[i].size, slabclass[i].perslab);
         }
     }
 
     power_largest = i;
-    slabclass[power_largest].size = POWER_BLOCK;
+    slabclass[power_largest].size = settings.item_size_max;
     slabclass[power_largest].perslab = 1;
+    if (settings.verbose > 1) {
+        fprintf(stderr, "slab class %3d: chunk size %9u perslab %7u\n",
+                i, slabclass[i].size, slabclass[i].perslab);
+    }
 
     /* for the test suite:  faking of how much we've already malloc'd */
     {
@@ -193,20 +196,16 @@ static int grow_slab_list (const unsigned int id) {
 
 static int do_slabs_newslab(const unsigned int id) {
     slabclass_t *p = &slabclass[id];
-#ifdef ALLOW_SLABS_REASSIGN
-    int len = POWER_BLOCK;
-#else
     int len = p->size * p->perslab;
-#endif
     char *ptr;
 
-    if (mem_limit && mem_malloced + len > mem_limit && p->slabs > 0)
+    if ((mem_limit && mem_malloced + len > mem_limit && p->slabs > 0) ||
+        (grow_slab_list(id) == 0) ||
+        ((ptr = memory_allocate((size_t)len)) == 0)) {
+
+        MEMCACHED_SLABS_SLABCLASS_ALLOCATE_FAILED(id);
         return 0;
-
-    if (grow_slab_list(id) == 0) return 0;
-
-    ptr = memory_allocate((size_t)len);
-    if (ptr == 0) return 0;
+    }
 
     memset(ptr, 0, (size_t)len);
     p->end_page_ptr = ptr;
@@ -214,50 +213,66 @@ static int do_slabs_newslab(const unsigned int id) {
 
     p->slab_list[p->slabs++] = ptr;
     mem_malloced += len;
+    MEMCACHED_SLABS_SLABCLASS_ALLOCATE(id);
+
     return 1;
 }
 
 /*@null@*/
-void *do_slabs_alloc(const size_t size, unsigned int id) {
+static void *do_slabs_alloc(const size_t size, unsigned int id) {
     slabclass_t *p;
+    void *ret = NULL;
 
-    if (id < POWER_SMALLEST || id > power_largest)
+    if (id < POWER_SMALLEST || id > power_largest) {
+        MEMCACHED_SLABS_ALLOCATE_FAILED(size, 0);
         return NULL;
+    }
 
     p = &slabclass[id];
     assert(p->sl_curr == 0 || ((item *)p->slots[p->sl_curr - 1])->slabs_clsid == 0);
 
 #ifdef USE_SYSTEM_MALLOC
-    if (mem_limit && mem_malloced + size > mem_limit)
+    if (mem_limit && mem_malloced + size > mem_limit) {
+        MEMCACHED_SLABS_ALLOCATE_FAILED(size, id);
         return 0;
+    }
     mem_malloced += size;
-    return malloc(size);
+    ret = malloc(size);
+    MEMCACHED_SLABS_ALLOCATE(size, id, 0, ret);
+    return ret;
 #endif
 
     /* fail unless we have space at the end of a recently allocated page,
        we have something on our freelist, or we could allocate a new page */
-    if (! (p->end_page_ptr != 0 || p->sl_curr != 0 || do_slabs_newslab(id) != 0))
-        return 0;
-
-    /* return off our freelist, if we have one */
-    if (p->sl_curr != 0)
-        return p->slots[--p->sl_curr];
-
-    /* if we recently allocated a whole page, return from that */
-    if (p->end_page_ptr) {
-        void *ptr = p->end_page_ptr;
+    if (! (p->end_page_ptr != 0 || p->sl_curr != 0 ||
+           do_slabs_newslab(id) != 0)) {
+        /* We don't have more memory available */
+        ret = NULL;
+    } else if (p->sl_curr != 0) {
+        /* return off our freelist */
+        ret = p->slots[--p->sl_curr];
+    } else {
+        /* if we recently allocated a whole page, return from that */
+        assert(p->end_page_ptr != NULL);
+        ret = p->end_page_ptr;
         if (--p->end_page_free != 0) {
-            p->end_page_ptr += p->size;
+            p->end_page_ptr = ((caddr_t)p->end_page_ptr) + p->size;
         } else {
             p->end_page_ptr = 0;
         }
-        return ptr;
     }
 
-    return NULL;  /* shouldn't ever get here */
+    if (ret) {
+        p->requested += size;
+        MEMCACHED_SLABS_ALLOCATE(size, id, p->size, ret);
+    } else {
+        MEMCACHED_SLABS_ALLOCATE_FAILED(size, id);
+    }
+
+    return ret;
 }
 
-void do_slabs_free(void *ptr, const size_t size, unsigned int id) {
+static void do_slabs_free(void *ptr, const size_t size, unsigned int id) {
     slabclass_t *p;
 
     assert(((item *)ptr)->slabs_clsid == 0);
@@ -265,6 +280,7 @@ void do_slabs_free(void *ptr, const size_t size, unsigned int id) {
     if (id < POWER_SMALLEST || id > power_largest)
         return;
 
+    MEMCACHED_SLABS_FREE(size, id, ptr);
     p = &slabclass[id];
 
 #ifdef USE_SYSTEM_MALLOC
@@ -282,115 +298,97 @@ void do_slabs_free(void *ptr, const size_t size, unsigned int id) {
         p->sl_total = new_size;
     }
     p->slots[p->sl_curr++] = ptr;
+    p->requested -= size;
     return;
 }
 
-/*@null@*/
-char* do_slabs_stats(int *buflen) {
-    int i, total;
-    char *buf = (char *)malloc(power_largest * 200 + 100);
-    char *bufcurr = buf;
+static int nz_strcmp(int nzlength, const char *nz, const char *z) {
+    int zlength=strlen(z);
+    return (zlength == nzlength) && (strncmp(nz, z, zlength) == 0) ? 0 : -1;
+}
 
-    *buflen = 0;
-    if (buf == NULL) return NULL;
+bool get_stats(const char *stat_type, int nkey, ADD_STAT add_stats, void *c) {
+    bool ret = true;
+
+    if (add_stats != NULL) {
+        if (!stat_type) {
+            /* prepare general statistics for the engine */
+            APPEND_STAT("bytes", "%llu", (unsigned long long)stats.curr_bytes);
+            APPEND_STAT("curr_items", "%u", stats.curr_items);
+            APPEND_STAT("total_items", "%u", stats.total_items);
+            APPEND_STAT("evictions", "%llu",
+                        (unsigned long long)stats.evictions);
+        } else if (nz_strcmp(nkey, stat_type, "items") == 0) {
+            item_stats(add_stats, c);
+        } else if (nz_strcmp(nkey, stat_type, "slabs") == 0) {
+            slabs_stats(add_stats, c);
+        } else if (nz_strcmp(nkey, stat_type, "sizes") == 0) {
+            item_stats_sizes(add_stats, c);
+        } else {
+            ret = false;
+        }
+    } else {
+        ret = false;
+    }
+
+    return ret;
+}
+
+/*@null@*/
+static void do_slabs_stats(ADD_STAT add_stats, void *c) {
+    int i, total;
+    /* Get the per-thread stats which contain some interesting aggregates */
+    struct thread_stats thread_stats;
+    threadlocal_stats_aggregate(&thread_stats);
 
     total = 0;
     for(i = POWER_SMALLEST; i <= power_largest; i++) {
         slabclass_t *p = &slabclass[i];
         if (p->slabs != 0) {
-            unsigned int perslab, slabs;
-
+            uint32_t perslab, slabs;
             slabs = p->slabs;
             perslab = p->perslab;
 
-            bufcurr += sprintf(bufcurr, "STAT %d:chunk_size %u\r\n", i, p->size);
-            bufcurr += sprintf(bufcurr, "STAT %d:chunks_per_page %u\r\n", i, perslab);
-            bufcurr += sprintf(bufcurr, "STAT %d:total_pages %u\r\n", i, slabs);
-            bufcurr += sprintf(bufcurr, "STAT %d:total_chunks %u\r\n", i, slabs*perslab);
-            bufcurr += sprintf(bufcurr, "STAT %d:used_chunks %u\r\n", i, slabs*perslab - p->sl_curr);
-            bufcurr += sprintf(bufcurr, "STAT %d:free_chunks %u\r\n", i, p->sl_curr);
-            bufcurr += sprintf(bufcurr, "STAT %d:free_chunks_end %u\r\n", i, p->end_page_free);
+            char key_str[STAT_KEY_LEN];
+            char val_str[STAT_VAL_LEN];
+            int klen = 0, vlen = 0;
+
+            APPEND_NUM_STAT(i, "chunk_size", "%u", p->size);
+            APPEND_NUM_STAT(i, "chunks_per_page", "%u", perslab);
+            APPEND_NUM_STAT(i, "total_pages", "%u", slabs);
+            APPEND_NUM_STAT(i, "total_chunks", "%u", slabs * perslab);
+            APPEND_NUM_STAT(i, "used_chunks", "%u",
+                            slabs*perslab - p->sl_curr - p->end_page_free);
+            APPEND_NUM_STAT(i, "free_chunks", "%u", p->sl_curr);
+            APPEND_NUM_STAT(i, "free_chunks_end", "%u", p->end_page_free);
+            APPEND_NUM_STAT(i, "mem_requested", "%llu",
+                            (unsigned long long)p->requested);
+            APPEND_NUM_STAT(i, "get_hits", "%llu",
+                    (unsigned long long)thread_stats.slab_stats[i].get_hits);
+            APPEND_NUM_STAT(i, "cmd_set", "%llu",
+                    (unsigned long long)thread_stats.slab_stats[i].set_cmds);
+            APPEND_NUM_STAT(i, "delete_hits", "%llu",
+                    (unsigned long long)thread_stats.slab_stats[i].delete_hits);
+            APPEND_NUM_STAT(i, "incr_hits", "%llu",
+                    (unsigned long long)thread_stats.slab_stats[i].incr_hits);
+            APPEND_NUM_STAT(i, "decr_hits", "%llu",
+                    (unsigned long long)thread_stats.slab_stats[i].decr_hits);
+            APPEND_NUM_STAT(i, "cas_hits", "%llu",
+                    (unsigned long long)thread_stats.slab_stats[i].cas_hits);
+            APPEND_NUM_STAT(i, "cas_badval", "%llu",
+                    (unsigned long long)thread_stats.slab_stats[i].cas_badval);
+
             total++;
         }
     }
-    bufcurr += sprintf(bufcurr, "STAT active_slabs %d\r\nSTAT total_malloced %llu\r\n", total, (unsigned long long)mem_malloced);
-    bufcurr += sprintf(bufcurr, "END\r\n");
-    *buflen = bufcurr - buf;
-    return buf;
+
+    /* add overall slab stats and append terminator */
+
+    APPEND_STAT("active_slabs", "%d", total);
+    APPEND_STAT("total_malloced", "%llu", (unsigned long long)mem_malloced);
+    add_stats(NULL, 0, NULL, 0, c);
 }
 
-#ifdef ALLOW_SLABS_REASSIGN
-/* Blows away all the items in a slab class and moves its slabs to another
-   class. This is only used by the "slabs reassign" command, for manual tweaking
-   of memory allocation. It's disabled by default since it requires that all
-   slabs be the same size (which can waste space for chunk size mantissas of
-   other than 2.0).
-   1 = success
-   0 = fail
-   -1 = tried. busy. send again shortly. */
-int do_slabs_reassign(unsigned char srcid, unsigned char dstid) {
-    void *slab, *slab_end;
-    slabclass_t *p, *dp;
-    void *iter;
-    bool was_busy = false;
-
-    if (srcid < POWER_SMALLEST || srcid > power_largest ||
-        dstid < POWER_SMALLEST || dstid > power_largest)
-        return 0;
-
-    p = &slabclass[srcid];
-    dp = &slabclass[dstid];
-
-    /* fail if src still populating, or no slab to give up in src */
-    if (p->end_page_ptr || ! p->slabs)
-        return 0;
-
-    /* fail if dst is still growing or we can't make room to hold its new one */
-    if (dp->end_page_ptr || ! grow_slab_list(dstid))
-        return 0;
-
-    if (p->killing == 0) p->killing = 1;
-
-    slab = p->slab_list[p->killing - 1];
-    slab_end = (char*)slab + POWER_BLOCK;
-
-    for (iter = slab; iter < slab_end; (char*)iter += p->size) {
-        item *it = (item *)iter;
-        if (it->slabs_clsid) {
-            if (it->refcount) was_busy = true;
-            item_unlink(it);
-        }
-    }
-
-    /* go through free list and discard items that are no longer part of this slab */
-    {
-        int fi;
-        for (fi = p->sl_curr - 1; fi >= 0; fi--) {
-            if (p->slots[fi] >= slab && p->slots[fi] < slab_end) {
-                p->sl_curr--;
-                if (p->sl_curr > fi) p->slots[fi] = p->slots[p->sl_curr];
-            }
-        }
-    }
-
-    if (was_busy) return -1;
-
-    /* if good, now move it to the dst slab class */
-    p->slab_list[p->killing - 1] = p->slab_list[p->slabs - 1];
-    p->slabs--;
-    p->killing = 0;
-    dp->slab_list[dp->slabs++] = slab;
-    dp->end_page_ptr = slab;
-    dp->end_page_free = dp->perslab;
-    /* this isn't too critical, but other parts of the code do asserts to
-       make sure this field is always 0.  */
-    for (iter = slab; iter < slab_end; (char*)iter += dp->size) {
-        ((item *)iter)->slabs_clsid = 0;
-    }
-    return 1;
-}
-#endif
-
 static void *memory_allocate(size_t size) {
     void *ret;
 
@@ -409,7 +407,7 @@ static void *memory_allocate(size_t size) {
             size += CHUNK_ALIGN_BYTES - (size % CHUNK_ALIGN_BYTES);
         }
 
-        mem_current += size;
+        mem_current = ((char*)mem_current) + size;
         if (size < mem_avail) {
             mem_avail -= size;
         } else {
@@ -419,3 +417,24 @@ static void *memory_allocate(size_t size) {
 
     return ret;
 }
+
+void *slabs_alloc(size_t size, unsigned int id) {
+    void *ret;
+
+    pthread_mutex_lock(&slabs_lock);
+    ret = do_slabs_alloc(size, id);
+    pthread_mutex_unlock(&slabs_lock);
+    return ret;
+}
+
+void slabs_free(void *ptr, size_t size, unsigned int id) {
+    pthread_mutex_lock(&slabs_lock);
+    do_slabs_free(ptr, size, id);
+    pthread_mutex_unlock(&slabs_lock);
+}
+
+void slabs_stats(ADD_STAT add_stats, void *c) {
+    pthread_mutex_lock(&slabs_lock);
+    do_slabs_stats(add_stats, c);
+    pthread_mutex_unlock(&slabs_lock);
+}
